"""
Module for adding the SpeX Prism Spectral Libraries to the database.
"""

import os
import urllib.request

import numpy as np
import pandas as pd

from astropy.io.votable import parse_single_table

from species.analysis import photometry
from species.util import data_util, query_util


def add_spex(input_path, database):
    """
    Function for adding the SpeX Prism Spectral Library to the database.

    Parameters
    ----------
    input_path : str
        Path of the data folder.
    database : h5py._hl.files.File
        The HDF5 database.

    Returns
    -------
    NoneType
        None
    """

    distance_url = 'https://home.strw.leidenuniv.nl/~stolker/species/distance.dat'
    distance_file = os.path.join(input_path, 'distance.dat')

    if not os.path.isfile(distance_file):
        urllib.request.urlretrieve(distance_url, distance_file)

    distance_data = pd.pandas.read_csv(distance_file,
                                       usecols=[0, 3, 4],
                                       names=['object', 'distance', 'distance_error'],
                                       delimiter=',',
                                       dtype={'object': str,
                                              'distance': float,
                                              'distance_error': float})

    database.create_group('spectra/spex')

    data_path = os.path.join(input_path, 'spex')

    if not os.path.exists(data_path):
        os.makedirs(data_path)

    url_all = 'http://svo2.cab.inta-csic.es/vocats/v2/spex/' \
              'cs.php?RA=180.000000&DEC=0.000000&SR=180.000000&VERB=2'

    xml_file_spex = os.path.join(data_path, 'spex.xml')

    if not os.path.isfile(xml_file_spex):
        urllib.request.urlretrieve(url_all, xml_file_spex)

    table = parse_single_table(xml_file_spex)
    # name = table.array['name']
    twomass = table.array['name2m']
    url = table.array['access_url']

    unique_id = []

    for i, item in enumerate(url):
        if twomass[i] not in unique_id:

            if isinstance(twomass[i], str):
                xml_file_1 = os.path.join(data_path, twomass[i]+'.xml')
            else:
                # Use decode for backward compatibility
                xml_file_1 = os.path.join(data_path, twomass[i].decode('utf-8')+'.xml')

            if not os.path.isfile(xml_file_1):
                if isinstance(item, str):
                    urllib.request.urlretrieve(item, xml_file_1)
                else:
                    urllib.request.urlretrieve(item.decode('utf-8'), xml_file_1)

            table = parse_single_table(xml_file_1)
            name = table.array['ID']
            url = table.array['access_url']

            if isinstance(name[0], str):
                name = name[0]
            else:
                name = name[0].decode('utf-8')

            print_message = f'Downloading SpeX Prism Spectral Library... {name}'
            print(f'\r{print_message:<72}', end='')

            xml_file_2 = os.path.join(data_path, f'spex_{name}.xml')

            if not os.path.isfile(xml_file_2):
                if isinstance(url[0], str):
                    urllib.request.urlretrieve(url[0], xml_file_2)
                else:
                    urllib.request.urlretrieve(url[0].decode('utf-8'), xml_file_2)

            unique_id.append(twomass[i])

    print_message = 'Downloading SpeX Prism Spectral Library... [DONE]'
    print(f'\r{print_message:<72}')

    h_twomass = photometry.SyntheticPhotometry('2MASS/2MASS.H')

    # 2MASS H band zero point for 0 mag (Cogen et al. 2003)
    h_zp = 1.133e-9  # (W m-2 um-1)

    for votable in os.listdir(data_path):
        if votable.startswith('spex_') and votable.endswith('.xml'):
            xml_file = os.path.join(data_path, votable)

            table = parse_single_table(xml_file)

            wavelength = table.array['wavelength']  # (Angstrom)
            flux = table.array['flux']  # Normalized units

            wavelength = np.array(wavelength*1e-4)  # (um)
            flux = np.array(flux)  # (a.u.)
            error = np.full(flux.size, np.nan)

            # 2MASS magnitudes
            j_mag = table.get_field_by_id('jmag').value
            h_mag = table.get_field_by_id('hmag').value
            ks_mag = table.get_field_by_id('ksmag').value

            if not isinstance(j_mag, str):
                j_mag = j_mag.decode('utf-8')

            if not isinstance(h_mag, str):
                h_mag = h_mag.decode('utf-8')

            if not isinstance(ks_mag, str):
                ks_mag = ks_mag.decode('utf-8')

            if j_mag == '':
                j_mag = np.nan
            else:
                j_mag = float(j_mag)

            if h_mag == '':
                h_mag = np.nan
            else:
                h_mag = float(h_mag)

            if ks_mag == '':
                ks_mag = np.nan
            else:
                ks_mag = float(ks_mag)

            name = table.get_field_by_id('name').value

            if not isinstance(name, str):
                name = name.decode('utf-8')

            twomass_id = table.get_field_by_id('name2m').value

            if not isinstance(twomass_id, str):
                twomass_id = twomass_id.decode('utf-8')

            try:
                sptype = table.get_field_by_id('nirspty').value

                if not isinstance(sptype, str):
                    sptype = sptype.decode('utf-8')

            except KeyError:
                try:
                    sptype = table.get_field_by_id('optspty').value

                    if not isinstance(sptype, str):
                        sptype = sptype.decode('utf-8')

                except KeyError:
                    sptype = 'None'

            sptype = data_util.update_sptype(np.array([sptype]))[0].strip()

            h_flux, _ = h_twomass.magnitude_to_flux(h_mag, error=None, zp_flux=h_zp)
            phot = h_twomass.spectrum_to_flux(wavelength, flux)  # Normalized units

            flux *= h_flux/phot[0]  # (W m-2 um-1)

            spdata = np.vstack([wavelength, flux, error])

            # simbad_id, distance = query_util.get_distance(f'2MASS {twomass_id}')
            simbad_id = query_util.get_simbad(f'2MASS {twomass_id}')

            if simbad_id is not None:
                if not isinstance(simbad_id, str):
                    simbad_id = simbad_id.decode('utf-8')

                dist_select = distance_data.loc[distance_data['object'] == simbad_id]

                if not dist_select.empty:
                    distance = (dist_select['distance'], dist_select['distance_error'])
                else:
                    distance = (np.nan, np.nan)

            else:
                distance = (np.nan, np.nan)

            if sptype[0] in ['M', 'L', 'T'] and len(sptype) == 2:
                print_message = f'Adding SpeX Prism Spectral Library... {name}'
                print(f'\r{print_message:<72}', end='')

                dset = database.create_dataset(f'spectra/spex/{name}', data=spdata)

                dset.attrs['name'] = str(name).encode()
                dset.attrs['sptype'] = str(sptype).encode()
                dset.attrs['simbad'] = str(simbad_id).encode()
                dset.attrs['2MASS/2MASS.J'] = j_mag
                dset.attrs['2MASS/2MASS.H'] = h_mag
                dset.attrs['2MASS/2MASS.Ks'] = ks_mag
                dset.attrs['distance'] = distance[0]  # (pc)
                dset.attrs['distance_error'] = distance[1]  # (pc)

    print_message = 'Adding SpeX Prism Spectral Library... [DONE]'
    print(f'\r{print_message:<72}')

    database.close()



"""
Module with reading functionalities for data from individual objects.
"""

import os
import configparser

from typing import Optional, Union, Tuple

import h5py
import numpy as np

from typeguard import typechecked

from species.util import phot_util


class ReadObject:
    """
    Class for reading data from an individual object from the database.
    """

    @typechecked
    def __init__(self,
                 object_name: str) -> None:
        """
        Parameters
        ----------
        object_name : str
            Object name as stored in the database (e.g. 'beta Pic b', 'PZ Tel B').

        Returns
        -------
        NoneType
            None
        """

        self.object_name = object_name

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

        with h5py.File(self.database, 'r') as h5_file:
            if f'objects/{self.object_name}' not in h5_file:
                raise ValueError(f'The object \'{self.object_name}\' is not present in the '
                                 f'database.')

    @typechecked
    def get_photometry(self,
                       filter_name: str) -> np.ndarray:
        """
        Function for reading the photometry of the object.

        Parameters
        ----------
        filter_name : str
            Filter ID.

        Returns
        -------
        np.ndarray
            Apparent magnitude, magnitude error (error), flux (W m-2 um-1),
            flux error (W m-2 um-1).
        """

        with h5py.File(self.database, 'r') as h5_file:
            if filter_name in h5_file[f'objects/{self.object_name}']:
                obj_phot = np.asarray(h5_file[f'objects/{self.object_name}/{filter_name}'])

            else:
                raise ValueError(f'There is no photometric data of {self.object_name} '
                                 f'available with the {filter_name} filter.')

        return obj_phot

    @typechecked
    def get_spectrum(self) -> dict:
        """
        Function for reading the spectra and covariance matrices of the object.

        Returns
        -------
        dict
            Dictionary with spectra and covariance matrices.
        """

        with h5py.File(self.database, 'r') as h5_file:
            if f'objects/{self.object_name}/spectrum' in h5_file:
                spectrum = {}

                for item in h5_file[f'objects/{self.object_name}/spectrum']:
                    data_group = f'objects/{self.object_name}/spectrum/{item}'

                    if f'{data_group}/covariance' not in h5_file:
                        spectrum[item] = (np.asarray(h5_file[f'{data_group}/spectrum']),
                                          None,
                                          None,
                                          h5_file[f'{data_group}'].attrs['specres'])

                    else:
                        spectrum[item] = (np.asarray(h5_file[f'{data_group}/spectrum']),
                                          np.asarray(h5_file[f'{data_group}/covariance']),
                                          np.asarray(h5_file[f'{data_group}/inv_covariance']),
                                          h5_file[f'{data_group}'].attrs['specres'])

            else:
                spectrum = None

        return spectrum

    @typechecked
    def get_distance(self) -> Tuple[float, float]:
        """
        Function for reading the distance to the object.

        Returns
        -------
        float
            Distance (pc).
        float
            Uncertainty (pc).
        """

        with h5py.File(self.database, 'r') as h5_file:
            obj_distance = np.asarray(h5_file[f'objects/{self.object_name}/distance'])

        return obj_distance[0], obj_distance[1]

    @typechecked
    def get_absmag(self,
                   filter_name: str) -> Union[Tuple[float, Optional[float]],
                                              Tuple[np.ndarray, Optional[np.ndarray]]]:
        """
        Function for calculating the absolute magnitudes of the object from the apparent
        magnitudes and distance. The errors on the apparent magnitude and distance are propagated
        into an error on the absolute magnitude.

        Parameters
        ----------
        filter_name : str
            Filter name.

        Returns
        -------
        float, np.ndarray
            Absolute magnitude.
        float, np.ndarray
            Error on the absolute magnitude.
        """

        with h5py.File(self.database, 'r') as h5_file:
            obj_distance = np.asarray(h5_file[f'objects/{self.object_name}/distance'])

            if filter_name in h5_file[f'objects/{self.object_name}']:
                obj_phot = np.asarray(h5_file[f'objects/{self.object_name}/{filter_name}'])

            else:
                raise ValueError(f'There is no photometric data of \'{self.object_name}\' '
                                 f'available with the {filter_name}.')

        if obj_phot.ndim == 1:
            abs_mag = phot_util.apparent_to_absolute((obj_phot[0], obj_phot[1]),
                                                     (obj_distance[0], obj_distance[1]))

        elif obj_phot.ndim == 2:
            abs_mag = phot_util.apparent_to_absolute((obj_phot[0, :], obj_phot[1, :]),
                                                     (obj_distance[0], obj_distance[1]))

        return abs_mag



import os
import shutil

import pytest
import numpy as np

import species
from species.util import test_util


class TestSpectrum:

    def setup_class(self):
        self.limit = 1e-8
        self.test_path = os.path.dirname(__file__) + '/'

    def teardown_class(self):
        os.remove('species_database.hdf5')
        os.remove('species_config.ini')
        os.remove('spectrum.pdf')
        shutil.rmtree('data/')

    def test_species_init(self):
        test_util.create_config('./')
        species.SpeciesInit()

    def test_read_spectrum(self):
        database = species.Database()

        with pytest.warns(UserWarning) as warning:
            database.add_spectrum('irtf', sptypes=['L', ])

        read_spectrum = species.ReadSpectrum('irtf', filter_name='MKO/NSFCam.H')
        assert read_spectrum.wavel_range == pytest.approx((1.382, 1.8656), rel=1e-6, abs=0.)

    def test_get_spectrum(self):
        read_spectrum = species.ReadSpectrum('irtf', filter_name='MKO/NSFCam.H')
        spec_box = read_spectrum.get_spectrum(sptypes=['L0', ], exclude_nan=True)

        assert spec_box.wavelength[0].shape == (1063, )
        assert spec_box.flux[0].shape == (1063, )

        assert np.sum(spec_box.wavelength[0]) == pytest.approx(1692.8604, rel=1e-7, abs=0.)
        assert np.sum(spec_box.flux[0]) == pytest.approx(4.5681937e-11, rel=1e-7, abs=0.)

        species.plot_spectrum(boxes=[spec_box, ],
                              filters=['MKO/NSFCam.H', ],
                              output='spectrum.pdf',
                              xlim=(1., 2.5),
                              offset=(-0.08, -0.06))

        assert os.path.exists('spectrum.pdf')



"""
Module with functionalities for calculating synthetic photometry.
"""

import os
import math
import warnings
import configparser

from typing import Optional, Union, Tuple, List

import h5py
import numpy as np

from typeguard import typechecked

from species.data import database
from species.read import read_filter, read_calibration
from species.util import phot_util


class SyntheticPhotometry:
    """
    Class for calculating synthetic photometry from a spectrum and also for conversion between
    magnitudes and fluxes. Note that depending on the detector type (energy- or photon-counting)
    the integral for the filter-weighted flux contains an additional wavelength factor.
    """

    @typechecked
    def __init__(self,
                 filter_name: str) -> None:
        """
        Parameters
        ----------
        filter_name : str
            Filter name as listed in the database. Filters from the SVO Filter Profile Service are
            automatically downloaded and added to the database.

        Returns
        -------
        NoneType
            None
        """

        self.filter_name = filter_name
        self.filter_interp = None
        self.wavel_range = None

        self.vega_mag = 0.03  # (mag)

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

        read_filt = read_filter.ReadFilter(self.filter_name)
        self.det_type = read_filt.detector_type()

    @typechecked
    def zero_point(self) -> np.float64:
        """
        Internal function for calculating the zero point of the provided ``filter_name``.

        Returns
        -------
        float
            Zero-point flux (W m-2 um-1).
        """

        if self.wavel_range is None:
            transmission = read_filter.ReadFilter(self.filter_name)
            self.wavel_range = transmission.wavelength_range()

        h5_file = h5py.File(self.database, 'r')

        try:
            h5_file['spectra/calibration/vega']

        except KeyError:
            h5_file.close()
            species_db = database.Database()
            species_db.add_spectrum('vega')
            h5_file = h5py.File(self.database, 'r')

        readcalib = read_calibration.ReadCalibration('vega', None)
        calibbox = readcalib.get_spectrum()

        wavelength = calibbox.wavelength
        flux = calibbox.flux

        wavelength_crop = wavelength[(wavelength > self.wavel_range[0]) &
                                     (wavelength < self.wavel_range[1])]

        flux_crop = flux[(wavelength > self.wavel_range[0]) &
                         (wavelength < self.wavel_range[1])]

        h5_file.close()

        return self.spectrum_to_flux(wavelength_crop, flux_crop)[0]

    @typechecked
    def spectrum_to_flux(self,
                         wavelength: np.ndarray,
                         flux: np.ndarray,
                         error: Optional[np.ndarray] = None,
                         threshold: Optional[float] = 0.05) -> Tuple[
                             Union[np.float32, np.float64],
                             Union[Optional[np.float32], Optional[np.float64]]]:
        """
        Function for calculating the average flux from a spectrum and a filter profile. The error
        is propagated by sampling 200 random values from the error distributions.

        Parameters
        ----------
        wavelength : np.ndarray
            Wavelength points (um).
        flux : np.ndarray
            Flux (W m-2 um-1).
        error : np.ndarray, None
            Uncertainty (W m-2 um-1). Not used if set to ``None``.
        threshold : float, None
            Transmission threshold (value between 0 and 1). If the minimum transmission value is
            larger than the threshold, a NaN is returned. This will happen if the input spectrum
            does not cover the full wavelength range of the filter profile. Not used if set to
            ``None``.

        Returns
        -------
        float
            Average flux (W m-2 um-1).
        float, None
            Uncertainty (W m-2 um-1).
        """

        if error is not None:
            # The error calculation requires the original spectrum because spectrum_to_flux is used
            wavel_error = wavelength.copy()
            flux_error = flux.copy()

        if self.filter_interp is None:
            transmission = read_filter.ReadFilter(self.filter_name)
            self.filter_interp = transmission.interpolate_filter()

            if self.wavel_range is None:
                self.wavel_range = transmission.wavelength_range()

        if wavelength.size == 0:
            raise ValueError(f'Calculation of the mean flux for {self.filter_name} is not '
                             f'possible because the wavelength array is empty.')

        indices = np.where((self.wavel_range[0] <= wavelength) &
                           (wavelength <= self.wavel_range[1]))[0]

        if indices.size < 2:
            syn_flux = np.nan

            warnings.warn('Calculating a synthetic flux requires more than one wavelength '
                          'point. Photometry is set to NaN.')

        else:
            if threshold is None and (wavelength[0] > self.wavel_range[0] or
                                      wavelength[-1] < self.wavel_range[1]):

                warnings.warn(f'The filter profile of {self.filter_name} '
                              f'({self.wavel_range[0]:.4f}-{self.wavel_range[1]:.4f}) extends '
                              f'beyond the wavelength range of the spectrum ({wavelength[0]:.4f} '
                              f'-{wavelength[-1]:.4f}). The flux is set to NaN. Setting the '
                              f'\'threshold\' parameter will loosen the wavelength constraints.')

                syn_flux = np.nan

            else:
                wavelength = wavelength[indices]
                flux = flux[indices]

                transmission = self.filter_interp(wavelength)

                if threshold is not None and \
                        (transmission[0] > threshold or transmission[-1] > threshold) and \
                        (wavelength[0] < self.wavel_range[0] or wavelength[-1] >
                         self.wavel_range[-1]):

                    warnings.warn(f'The filter profile of {self.filter_name} '
                                  f'({self.wavel_range[0]:.4f}-{self.wavel_range[1]:.4f}) '
                                  f'extends beyond the wavelength range of the spectrum '
                                  f'({wavelength[0]:.4f}-{wavelength[-1]:.4f}). The flux '
                                  f'is set to NaN. Increasing the \'threshold\' parameter '
                                  f'({threshold}) will loosen the wavelength constraint.')

                    syn_flux = np.nan

                else:
                    indices = np.isnan(transmission)
                    indices = np.logical_not(indices)

                    if self.det_type == 'energy':
                        # Energy counting detector
                        integrand1 = transmission[indices]*flux[indices]
                        integrand2 = transmission[indices]

                    if self.det_type == 'photon':
                        # Photon counting detector
                        integrand1 = wavelength[indices]*transmission[indices]*flux[indices]
                        integrand2 = wavelength[indices]*transmission[indices]

                    integral1 = np.trapz(integrand1, wavelength[indices])
                    integral2 = np.trapz(integrand2, wavelength[indices])

                    syn_flux = integral1/integral2

        if error is not None and not np.any(np.isnan(error)):
            phot_random = np.zeros(200)

            for i in range(200):
                # Use the original spectrum size (i.e. wavel_error and flux_error)
                spec_random = flux_error + np.random.normal(loc=0.,
                                                            scale=1.,
                                                            size=wavel_error.shape[0])*error

                phot_random[i] = self.spectrum_to_flux(wavel_error,
                                                       spec_random,
                                                       error=None,
                                                       threshold=threshold)[0]

            error_flux = np.std(phot_random)

        else:
            error_flux = None

        return syn_flux, error_flux

    @typechecked
    def spectrum_to_magnitude(self,
                              wavelength: np.ndarray,
                              flux: np.ndarray,
                              error: Optional[Union[np.ndarray, List[np.ndarray]]] = None,
                              distance: Optional[Tuple[float, Optional[float]]] = None,
                              threshold: Optional[float] = 0.05) -> Tuple[
                                  Tuple[float, Optional[float]],
                                  Optional[Tuple[Optional[float], Optional[float]]]]:
        """
        Function for calculating the apparent and absolute magnitude from a spectrum and a
        filter profile. The error is propagated by sampling 200 random values from the error
        distributions.

        Parameters
        ----------
        wavelength : np.ndarray
            Wavelength points (um).
        flux : np.ndarray
            Flux (W m-2 um-1).
        error : np.ndarray, list(np.ndarray), None
            Uncertainty (W m-2 um-1).
        distance : tuple(float, float), None
            Distance and uncertainty (pc). No absolute magnitude is calculated if set to ``None``.
            No error on the absolute magnitude is calculated if the uncertainty is set to ``None``.
        threshold : float, None
            Transmission threshold (value between 0 and 1). If the minimum transmission value is
            larger than the threshold, a NaN is returned. This will happen if the input spectrum
            does not cover the full wavelength range of the filter profile. Not used if set to
            ``None``.

        Returns
        -------
        tuple(float, float)
            Apparent magnitude and uncertainty.
        tuple(float, float)
            Absolute magnitude and uncertainty.
        """

        zp_flux = self.zero_point()

        syn_flux = self.spectrum_to_flux(wavelength,
                                         flux,
                                         error=error,
                                         threshold=threshold)

        app_mag = self.vega_mag - 2.5*math.log10(syn_flux[0]/zp_flux)

        if error is not None and not np.any(np.isnan(error)):
            mag_random = np.zeros(200)

            for i in range(200):
                spec_random = flux + np.random.normal(loc=0.,
                                                      scale=1.,
                                                      size=wavelength.shape[0])*error

                flux_random = self.spectrum_to_flux(wavelength,
                                                    spec_random,
                                                    error=None,
                                                    threshold=threshold)

                mag_random[i] = self.vega_mag - 2.5*np.log10(flux_random[0]/zp_flux)

            error_app_mag = np.std(mag_random)

        else:
            error_app_mag = None

        if distance is None:
            abs_mag = None
            error_abs_mag = None

        else:
            abs_mag = app_mag - 5.*np.log10(distance[0]) + 5.

            if error_app_mag is not None and distance[1] is not None:
                error_dist = distance[1] * (5./(distance[0]*math.log(10.)))
                error_abs_mag = math.sqrt(error_app_mag**2 + error_dist**2)

            else:
                error_abs_mag = None

        return (app_mag, error_app_mag), (abs_mag, error_abs_mag)

    @typechecked
    def magnitude_to_flux(self,
                          magnitude: float,
                          error: Optional[float] = None,
                          zp_flux: Optional[float] = None) -> Tuple[np.float64, np.float64]:
        """
        Function for converting a magnitude to a flux.

        Parameters
        ----------
        magnitude : float
            Magnitude.
        error : float, None
            Error on the magnitude. Not used if set to ``None``.
        zp_flux : float, None
            Zero-point flux (W m-2 um-1). The value is calculated if set to ``None``.

        Returns
        -------
        float
            Flux (W m-2 um-1).
        float
            Error (W m-2 um-1).
        """

        if zp_flux is None:
            zp_flux = self.zero_point()

        flux = 10.**(-0.4*(magnitude-self.vega_mag))*zp_flux

        if error is None:
            error_flux = None

        else:
            error_upper = flux * (10.**(0.4*error) - 1.)
            error_lower = flux * (1. - 10.**(-0.4*error))
            error_flux = (error_lower+error_upper)/2.

        return flux, error_flux

    @typechecked
    def flux_to_magnitude(self,
                          flux: float,
                          error: Optional[Union[float, np.ndarray]] = None,
                          distance: Optional[Union[Tuple[float, Optional[float]],
                                                   Tuple[np.ndarray, Optional[np.ndarray]]]] = None
                          ) -> Tuple[Union[Tuple[float, Optional[float]],
                                           Tuple[np.ndarray, Optional[np.ndarray]]],
                                     Union[Tuple[float, Optional[float]],
                                           Tuple[np.ndarray, Optional[np.ndarray]]]]:
        """
        Function for converting a flux into a magnitude.

        Parameters
        ----------
        flux : float, np.ndarray
            Flux (W m-2 um-1).
        error : float, np.ndarray, None
            Uncertainty (W m-2 um-1). Not used if set to None.
        distance : tuple(float, float), tuple(np.ndarray, np.ndarray)
            Distance and uncertainty (pc). The returned absolute magnitude is set to None in case
            ``distance`` is set to None. The error is not propagated into the error on the absolute
            magnitude in case the distance uncertainty is set to None, for example
            ``distance=(20., None)``

        Returns
        -------
        tuple(float, float), tuple(np.ndarray, np.ndarray)
            Apparent magnitude and uncertainty.
        tuple(float, float), tuple(np.ndarray, np.ndarray)
            Absolute magnitude and uncertainty.
        """

        zp_flux = self.zero_point()

        app_mag = self.vega_mag - 2.5*np.log10(flux/zp_flux)

        if error is None:
            error_app_mag = None
            error_abs_mag = None

        else:
            error_app_lower = app_mag - (self.vega_mag - 2.5*np.log10((flux+error)/zp_flux))
            error_app_upper = (self.vega_mag - 2.5*np.log10((flux-error)/zp_flux)) - app_mag
            error_app_mag = (error_app_lower+error_app_upper)/2.

        if distance is None:
            abs_mag = None
            error_abs_mag = None

        else:
            abs_mag, error_abs_mag = phot_util.apparent_to_absolute(
                (app_mag, error_app_mag), distance)

        return (app_mag, error_app_mag), (abs_mag, error_abs_mag)



from species.analysis.fit_model import FitModel

from species.analysis.fit_spectrum import FitSpectrum

from species.analysis.photometry import SyntheticPhotometry

# from species.analysis.retrieval import AtmosphericRetrieval

from species.read.read_calibration import ReadCalibration

from species.read.read_filter import ReadFilter

from species.read.read_isochrone import ReadIsochrone

from species.read.read_model import ReadModel

from species.read.read_planck import ReadPlanck

# from species.read.read_radtrans import ReadRadtrans

from species.read.read_spectrum import ReadSpectrum

from species.read.read_color import ReadColorMagnitude, \
                                    ReadColorColor

from species.read.read_object import ReadObject

from species.core.box import create_box

from species.core.constants import *

from species.core.setup import SpeciesInit

from species.data.companions import get_data

from species.data.database import Database

from species.plot.plot_color import plot_color_magnitude, \
                                    plot_color_color

from species.plot.plot_mcmc import plot_posterior, \
                                   plot_walkers, \
                                   plot_mag_posterior, \
                                   plot_size_distributions, \
                                   plot_extinction

# from species.plot.plot_retrieval import plot_pt_profile

from species.plot.plot_spectrum import plot_spectrum

from species.util.phot_util import apparent_to_absolute, \
                                   absolute_to_apparent, \
                                   multi_photometry, \
                                   get_residuals

from species.util.query_util import get_parallax

from species.util.read_util import add_luminosity, \
                                   get_mass, \
                                   powerlaw_spectrum, \
                                   update_spectra
                                   

__author__ = 'Tomas Stolker'
__license__ = 'MIT'
__version__ = '0.3.3'
__maintainer__ = 'Tomas Stolker'
__email__ = 'stolker@strw.leidenuniv.nl'
__status__ = 'Development'
